// OpenCP Module Player
// copyright (c) '94-'98 Niklas Beisert <nbeisert@physik.tu-muenchen.de>
//
// Player device for MPEG layer 3 file output
//
// revision history: (please note changes here)
//  -nb980510   Niklas Beisert <nbeisert@physik.tu-muenchen.de>
//    -first release
//  -kb981118   Tammo Hinrichs <opencp@gmx.net>
//    -reduced max buffer size to 32k to avoid problems with the
//     wave/mp3 player
//  -fd981206   Felix Domke    <tmbinc@gmx.net>
//    -edited for new binfile
//  -doj20020414 Dirk Jagdmann <doj@cubic.org>
//    -now using lame engine for mp3 output
//  -doj20020901 Dirk Jagdmann <doj@cubic.org>
//    -removed pwProccessKey()

#include <string.h>

#include "binfstd.h"
#include "imsdev.h"
#include "imsrtns.h"
#include "lame.h"
#include "player.h"
#include "pmain.h"

extern "C" extern sounddevice plrMPxWriter;

static long filepos=0;
static unsigned long buflen=0;
static char *playbuf=0;
static unsigned long bufpos=0;
static unsigned long bufrate=0;
static sbinfile file;
static unsigned char *diskcache=0;
static unsigned long cachelen=0;
static unsigned long cachepos=0;
static char busy=0;

static unsigned char *outputbuf=0;
static int outputbufsize=0;

static unsigned short samplerate;
static unsigned char stereo;

static lame_global_flags *lameflags=0;

static void Flush()
{
  busy=1;
  if (cachepos>(cachelen-bufrate*2))
  {
    int r=cachepos/sizeof(short)/(stereo?2:1);
    if((r=lame_encode_buffer_interleaved(lameflags, (short*)diskcache, r, outputbuf, outputbufsize)) < 0)
      {
	DEBUGSTR("could not encode frame");
      }
    else if(r==0)
      {
	DEBUGSTR("no mp3 bytes encoded");
      }
    else
      {
	if(!file.ewrite(outputbuf, r))
	  {
	    DEBUGSTR("could not wrote mp3 frames");
	  }
      }

    filepos+=cachepos;
    cachepos=0;
  }
  busy=0;
}

static int getpos()
{
  if (busy||((cachepos+bufrate)>cachelen))
    return bufpos;

  return (bufpos+bufrate)%buflen;
}

static void advance(int pos)
{
  int n=(pos-bufpos+buflen)%buflen;

  if ((bufpos+n)>buflen)
  {
    memcpy(diskcache+cachepos, playbuf+bufpos, buflen-bufpos-n);
    memcpy(diskcache+cachepos+buflen-bufpos-n, playbuf, n-buflen+bufpos+n);
  }
  else
    memcpy(diskcache+cachepos, playbuf+bufpos, n);
  cachepos+=n;

  bufpos=pos;
}

static long gettimer()
{
  return imuldiv(filepos+cachepos, 65536>>(stereo+1), samplerate);
}

static void dwSetOptions(int rate, int opt)
{
  stereo=!!(opt&PLR_STEREO);
  opt&=0xFFFF;

  opt|=PLR_16BIT|PLR_SIGNEDOUT;

  samplerate=rate;
  plrRate=rate;
  plrOpt=opt;
}

static int dwPlay(void *&buf, int &len)
{
  // lame init
  lameflags=lame_init();
  if(lameflags==0)
    {
      DEBUGSTR("could not init lame");
      return 0;
    }

  // decode VBR setting from bitrate
  int vbr=!!(savebitrate()&0x10000);
  // remove VBR setting from bitrate
  savebitrate(savebitrate()&0xFFFF);

  lame_set_quality(lameflags, 0);
  lame_set_in_samplerate(lameflags, samplerate);
  lame_set_num_channels(lameflags, stereo?2:1);
  lame_set_brate(lameflags, savebitrate());
  lame_set_VBR(lameflags, vbr?vbr_default:vbr_off);

  // all values should be set now...
  if(lame_init_params(lameflags) < 0)
    {
      DEBUGSTR("could not init lame parameters");
      return 0;
    }
  // lame init finished

  if (len>32704)
    len=32704;
  if(playbuf)
    delete [] playbuf;
  buf=playbuf=new char [len];
  if (!buf)
    return 0;
  memsetd(buf, (plrOpt&PLR_SIGNEDOUT)?0:(plrOpt&PLR_16BIT)?0x80008000:0x80808080, len>>2);

  cachelen=(1024*16*samplerate/44100)<<stereo;
  cachepos=0;
  if(diskcache)
    delete [] diskcache;
  diskcache=new unsigned char [cachelen];
  if (!diskcache)
    return 0;

  outputbufsize=(int)(1.2*cachelen+7200);
  if(outputbuf)
    delete [] outputbuf;
  outputbuf=new unsigned char[outputbufsize];
  if(!outputbuf)
    return 0;

  char fn[15];
  int i;
  for (i=0; i<1000; i++)
  {
    memcpy(fn, "CPOUT000.MP3", 13);
    fn[5]+=(i/100)%10;
    fn[6]+=(i/10)%10;
    fn[7]+=(i/1)%10;
    if (file.open(fn, file.openro))
      break;
  }

  file.open(fn, file.opentr|file.openos);

  buflen=len;
  bufpos=0;
  busy=0;
  bufrate=buflen/2;
  filepos=0;

  plrGetBufPos=getpos;
  plrGetPlayPos=getpos;
  plrAdvanceTo=advance;
  plrIdle=Flush;
  plrGetTimer=gettimer;

  return 1;
}

static void dwStop()
{
  plrIdle=0;

  int r;
  if((r=lame_encode_flush(lameflags, diskcache, cachelen)) < 0)
    {
      DEBUGSTR("could not flush encoding buffers");
    }
  else
    file.write(diskcache, r);

  if(lame_close(lameflags) < 0)
    {
      DEBUGSTR("could not close lame");
    }

  file.close();
  if(playbuf)
    delete [] playbuf;
  playbuf=0;
  if(diskcache)
    delete [] diskcache;
  diskcache=0;
  if(outputbuf)
    delete [] outputbuf;
  outputbuf=0;
}

static int dwInit(const deviceinfo &d)
{
  plrSetOptions=dwSetOptions;
  plrPlay=dwPlay;
  plrStop=dwStop;

  return 1;
}

static void dwClose()
{
  plrPlay=0;
}

static int dwDetect(deviceinfo &card)
{
  card.dev=&plrMPxWriter;
  card.port=-1;
  card.port2=-1;
  card.irq=-1;
  card.irq2=-1;
  card.dma=-1;
  card.dma2=-1;
  card.subtype=-1;
  card.mem=0;
  card.chan=2;

  return 1;
}


#include "devigen.h"
#include "psetting.h"

static unsigned long dwGetOpt(const char *sec)
{
  unsigned long opt=0;

  int mp3bitrate=cfGetProfileInt(sec, "bitrate", 0, 10);
  if (mp3bitrate>=1000)
    mp3bitrate/=1000;
  if(mp3bitrate>320)
    mp3bitrate=320;

  // encode VBR setting into bitrate
  mp3bitrate|=cfGetProfileBool(sec, "vbr", 0, 0)?0x10000:0;

  savebitrate(mp3bitrate);

  opt|=PLR_STEREO;

  return opt;
}

extern "C"
{
  sounddevice plrMPxWriter={SS_PLAYER, "MP3 Writer", dwDetect, dwInit, dwClose};
  devaddstruct plrMPxAdd = {dwGetOpt, 0, 0, 0};
  char *dllinfo = "driver _plrMPxWriter; addprocs _plrMPxAdd";
}
